import com.mashape.unirest.http.Unirest;
import com.mashape.unirest.http.exceptions.UnirestException;
import org.json.JSONObject;

import javax.crypto.Cipher;
import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.io.*;
import java.nio.charset.StandardCharsets;
import java.security.*;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.*;
import java.sql.Timestamp;

public class Demo {
    static String safecode = "8b998abba49ae33a194cb9677b13ecbe";
    static String requestUrl = "http://service.local.test";
    static String userId = "1";
    static String requestType = "payment_v2"; // payment_v2, withdraw_v2, payment_query_v2, withdraw_query_v2, balance_v2
    static String transcationSerect = "f55c19747fde060a4af40dc6ed187ae3";
    static String guid = "smajhv5r2etr";

    public static void main(String[] args) throws UnirestException {

        Map<String, String> params = new HashMap<>();
        Timestamp timestamp = new Timestamp(System.currentTimeMillis());

        params.put("user_id", userId);
        params.put("order_id", "TEST" + userId + timestamp.getTime());
        params.put("amount", String.valueOf(Math.floor(Math.random() * (1000 - 100 + 1) + 100)));
        params.put("currency", "PHP");
        params.put("channel", "QRPH");
        params.put("callback_url", "http://my.service/callback");
        params.put("redirect_url", "https://my.service/redirect");
        params.put("timestamp", String.valueOf((int) Math.floor(timestamp.getTime() / 1000)));
        params.put("remark", "test");
        params.put("email", "aaa@aaa.aaa");

        try {
            JSONObject jsonBody = sendRequest(requestType, params);
            int code = jsonBody.getInt("code");
            if (code != 1000) {
                throw new Exception(jsonBody.getString("message"));
            }
        } catch (Exception e) {
            System.out.println("Error:: " + e.getMessage());
        }
    }

    public static String generateSignature(Map<String, String> params, String signType) {
        String serializeSign = serializeSign(params, signType);

        System.out.println("generateSignature serializeSign::" + serializeSign);

        try {
            Signature privateSignature = Signature.getInstance("SHA256withRSA");
            PrivateKey privateKey = getPrivateKey("src/main/resources/private_key_pkcs8.pem");
            privateSignature.initSign(privateKey);
            privateSignature.update(serializeSign.getBytes("UTF-8"));
            byte[] sign = privateSignature.sign();
            return Base64.getEncoder().encodeToString(sign);
        } catch (Exception e) {
            e.printStackTrace();
        }

        return null;
    }

    public static String aes256Encode(String str, String code) {
        byte[] key = code.getBytes();
        byte[] result = null;
        try {
            Cipher cipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
            SecretKeySpec keySpec = new SecretKeySpec(key, "AES");
            cipher.init(Cipher.ENCRYPT_MODE, keySpec);
            result = cipher.doFinal(str.getBytes("UTF-8"));
        } catch (Exception e) {
            e.printStackTrace();
        }
        return Base64.getEncoder().encodeToString(result);
    }

    public static PublicKey getPublicKey(String filePath) {
        try {
            File file = new File(filePath);
            FileInputStream fis = new FileInputStream(file);
            DataInputStream dis = new DataInputStream(fis);
            byte[] keyBytes = new byte[(int) file.length()];
            dis.readFully(keyBytes);
            dis.close();

            String temp = new String(keyBytes);
            String publicKeyPEM = temp.replace("-----BEGIN PUBLIC KEY-----\n", "");
            publicKeyPEM = publicKeyPEM.replace("-----END PUBLIC KEY-----", "");
            byte[] decoded = Base64.getDecoder().decode(publicKeyPEM.replace("\n", ""));

            X509EncodedKeySpec spec = new X509EncodedKeySpec(decoded);
            KeyFactory kf = KeyFactory.getInstance("RSA");
            return kf.generatePublic(spec);
        } catch (Exception e) {
            e.printStackTrace();
        }

        return null;
    }

    public static PrivateKey getPrivateKey(String filePath) {
        try {
            File file = new File(filePath);
            FileInputStream fis = new FileInputStream(file);
            DataInputStream dis = new DataInputStream(fis);
            byte[] keyBytes = new byte[(int) file.length()];
            dis.readFully(keyBytes);
            dis.close();

            String temp = new String(keyBytes);
            String privateKeyPEM = temp.replace("-----BEGIN PRIVATE KEY-----\n", "");
            privateKeyPEM = privateKeyPEM.replace("-----END PRIVATE KEY-----", "");
            byte[] decoded = Base64.getDecoder().decode(privateKeyPEM.replace("\n", ""));

            PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec(decoded);
            KeyFactory kf = KeyFactory.getInstance("RSA");

            return kf.generatePrivate(privateKeySpec);
        } catch (Exception e) {
            e.printStackTrace();
        }

        return null;
    }

    public static JSONObject sendRequest(String requestType, Map<String, String> params)
            throws UnirestException {
        JSONObject jsonObject = new JSONObject();
        SortedSet<String> parameter = new TreeSet<>(params.keySet());
        for (String key : parameter) {
            jsonObject.put(key, params.get(key));
        }

        jsonObject.put("sign", generateSignature(params, requestType));
        System.out.println("json:: " + jsonObject.toString());
        String aesString = aes256Encode(jsonObject.toString(), safecode);
        Unirest.setTimeouts(0, 0);
        System.out.println("AES:: " + aesString);
        System.out.println("transactionSignature:: " + transactionSignature(aesString, transcationSerect));

        var jsonResponse = Unirest.post(requestUrl + getRoute(requestType))
                .header("Content-Type", "text/plain")
                .header("X-Transaction-Signature", transactionSignature(aesString, transcationSerect))
                .body(aesString).asJson();

        System.out.println("sendRequest response:: " + jsonResponse.getBody());
        return jsonResponse.getBody().getObject();
    }

    public static boolean verifySignature(Map<String, String> params, PublicKey publicKey, String sign,
            String signType) {
        try {
            String serializeSign = serializeSign(params, signType);
            Signature publicSignature = Signature.getInstance("SHA256withRSA");
            publicSignature.initVerify(publicKey);
            publicSignature.update(serializeSign.getBytes("UTF-8"));
            byte[] signDecoded = Base64.getDecoder().decode(sign.replace("\n", ""));
            boolean verifies = publicSignature.verify(signDecoded);
            return verifies;
        } catch (Exception e) {
            e.printStackTrace();
        }

        return false;
    }

    public static String serializeSign(Map<String, String> params, String signType) {
        String[] signFields;
        Map<String, String> signParams = new HashMap<>();

        switch (signType) {
            case "payment_v2":
            case "withdraw_v2":
                signFields = new String[] {
                        "user_id", "order_id", "amount", "currency", "channel",
                        "timestamp",
                };
                break;
            case "payment_query_v2":
            case "withdraw_query_v2":
                signFields = new String[] { "user_id", "order_id", "timestamp" };
                break;
            case "balance_v2":
                signFields = new String[] { "user_id", "timestamp" };
                break;

            case "payment_response":
                signFields = new String[] { "user_id", "order_id", "transaction_id", "channel",
                        "submit_currency", "submit_amount", "accept_currency", "accept_amount",
                        "exchange_rate", "pay_url" };
                break;

            case "withdraw_response":
                signFields = new String[] { "user_id", "order_id", "transaction_id", "channel",
                        "submit_currency", "submit_amount", "accept_currency", "accept_amount",
                        "exchange_rate" };
                break;

            case "payment_query_response":
            case "withdraw_query_response":
                signFields = new String[] { "user_id", "order_id", "transaction_id", "channel",
                        "submit_currency", "submit_amount", "accept_currency", "accept_amount",
                        "exchange_rate", "status", "timestamp" };
                break;

            case "balance_response":

            default:
                signFields = new String[] { "user_id" };
                break;
        }

        if (signFields.length > 0) {
            for (String key : signFields) {
                if (params.containsKey(key)) {
                    signParams.put(key, params.get(key));
                }
            }
        }

        Map<String, String> sortedParams = new TreeMap<>(signParams);

        StringBuilder sign = new StringBuilder();
        for (Map.Entry<String, String> entry : sortedParams.entrySet()) {
            sign.append(entry.getKey()).append('=').append(entry.getValue()).append('&');
        }
        sign.append(safecode);

        return sign.toString();
    }

    public static String transactionSignature(String requestBody, String transactionSecret) {
        try {
            // 使用 HmacSHA256 簽名算法
            Mac sha256_HMAC = Mac.getInstance("HmacSHA256");
            SecretKeySpec secretKey = new SecretKeySpec(transactionSecret.getBytes(StandardCharsets.UTF_8),
                    "HmacSHA256");
            sha256_HMAC.init(secretKey);

            // 計算 HMAC 散列
            byte[] hashBytes = sha256_HMAC.doFinal(requestBody.getBytes(StandardCharsets.UTF_8));

            // 使用 Base64 編碼
            String signature = Base64.getEncoder().encodeToString(hashBytes);

            return signature;
        } catch (Exception e) {
            e.printStackTrace();
            return null; // 適當的錯誤處理
        }
    }

    public static String getRoute(String type) {
        Map<String, String> routes = new HashMap<>();
        routes.put("payment_v2", String.format("/merchant/v2/%s/payment", guid));
        routes.put("withdraw_v2", String.format("/merchant/v2/%s/withdraw", guid));
        routes.put("payment_query_v2", String.format("/merchant/v2/%s/order/payment", guid));
        routes.put("withdraw_query_v2", String.format("/merchant/v2/%s/order/withdraw", guid));
        routes.put("balance_v2", String.format("/merchant/v2/%s/balance", guid));

        return routes.get(type);
    }
}